#!/bin/sh

function die {
   echo $1
   exit 1
}

CONFIGURE_ARGS=$*

(echo "${CONFIGURE_ARGS}" | grep -q -- '--enable-all-statics') && {
  BUILDING_STATIC="-static"
}

PACKAGE_NAME="fwbuilder${BUILDING_STATIC}"

RPMROOT=/var/tmp/${PACKAGE_NAME}
EXPORTDIR=`pwd`/export
SPECFILE=fwbuilder.spec

export CCACHE_DIR=$HOME/.ccache

./autogen.sh --prefix=/usr ${CONFIGURE_ARGS} || exit 1


eval `egrep '^VERSION|FWB_.*_VERSION' configure.in  | egrep -v 'AC_|AM_'`

TAR_FILE=${PACKAGE_NAME}-${VERSION}.tar.gz
SNAPSHOT_DIR=${PACKAGE_NAME}-${VERSION}

./maketar.sh ${PACKAGE_NAME} $VERSION

echo " === Building RPM for ${PACKAGE_NAME} version $VERSION === "

mkdir -p ${RPMROOT}
mkdir -p ${RPMROOT}/SOURCES
mkdir -p ${RPMROOT}/BUILD
mkdir -p ${RPMROOT}/SPECS
mkdir -p ${RPMROOT}/SRCS
mkdir -p ${RPMROOT}/RPMS
mkdir -p ${RPMROOT}/RPMS/i386
mkdir -p ${RPMROOT}/SRPMS

if [ ! -d $EXPORTDIR ]; then
    echo "Can't find export directory"
    exit 1
else

    if [ ! -f ${EXPORTDIR}/${TAR_FILE} ]; then
	make tar
    fi

    cp ${EXPORTDIR}/${TAR_FILE} ${RPMROOT}/SOURCES
    cp ${SPECFILE} ${RPMROOT}/SPECS 
    cd ${RPMROOT}/SPECS

    DISTRO=`cat /etc/issue|grep release|sed 's/Linux.*$//; s/ //g'`
    if test "x${DISTRO}" == "xMandrake"; then
        ARCH=`uname -m`
        if [ "X$ARCH" = "Xi686" ]; then
                RPMARCH="--target i586"
        fi
    fi

    rm -f .rpmmacros
    echo "%_topdir ${RPMROOT}" > .rpmmacros
    echo "%_unpackaged_files_terminate_build 0" >> .rpmmacros

    export HOME=`pwd`; rpmbuild -ba $RPMARCH ${SPECFILE} 

fi

