/* 

                          Firewall Builder

                 Copyright (C) 2002 NetCitadel, LLC

  Author:  Vadim Kurland     vadim@vk.crocodile.org

  $Id: NATCompiler_ipt_writers.cc,v 1.21 2002/12/13 08:05:09 vkurland Exp $

  This program is free software which we release under the GNU General Public
  License. You may redistribute and/or modify this program under the terms
  of that license as published by the Free Software Foundation; either
  version 2 of the License, or (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
 
  To get a copy of the GNU General Public License, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

*/

#include "NATCompiler_ipt.hh"

#include "fwbuilder/AddressRange.hh"
#include "fwbuilder/RuleElement.hh"
#include "fwbuilder/IPService.hh"
#include "fwbuilder/ICMPService.hh"
#include "fwbuilder/TCPService.hh"
#include "fwbuilder/UDPService.hh"
#include "fwbuilder/CustomService.hh"
#include "fwbuilder/Policy.hh"
#include "fwbuilder/Interface.hh"
#include "fwbuilder/IPv4.hh"
#include "fwbuilder/Firewall.hh"

#include <iostream>
#if __GNUC__ > 3 || \
    (__GNUC__ == 3 && (__GNUC_MINOR__ > 2 || (__GNUC_MINOR__ == 2 ) ) )
#  include <streambuf>
#else
#  include <streambuf.h>
#endif
#include <iomanip>
#include <fstream>
#include <sstream>

#include <assert.h>

using namespace libfwbuilder;
using namespace fwcompiler;
using namespace std;



/**
 *-----------------------------------------------------------------------
 *                    Methods for printing
 */

void NATCompiler_ipt::PrintRule::_printChainDirectionAndInterface(NATRule *rule)
{
    string       iface_id = rule->getInterfaceId();
    Interface *rule_iface = compiler->getCachedFwInterface(iface_id);
    string iface_name= (rule_iface!=NULL) ? rule_iface->getName() : "";

    compiler->output << "-A " << rule->getStr("ipt_chain") << " ";

    switch (rule->getRuleType()) {
    case NATRule::SNAT:  
	if (!iface_name.empty()) compiler->output << "-o " << iface_name;
	break;
    case NATRule::Masq:  
	if (!iface_name.empty()) compiler->output << "-o " << iface_name;
	break;
    case NATRule::DNAT:  
	if (!iface_name.empty()) compiler->output << "-i " << iface_name;
	break;
    case NATRule::Redirect: 
	if (!iface_name.empty()) compiler->output << "-i " << iface_name;
	break;
    }
    compiler->output << " ";
}



void NATCompiler_ipt::PrintRule::_printProtocol(Service *srv)
{
    if ( ! srv->isAny() && ! CustomService::isA(srv) ) {
        compiler->output << "-p " <<  srv->getProtocolName() << " ";
    }
}

void NATCompiler_ipt::PrintRule::_printMultiport(NATRule *rule)
{
    if( rule->getBool("ipt_multiport"))
	compiler->output << "-m multiport ";
}


string NATCompiler_ipt::PrintRule::_printOPorts(int rs,int re)
{
    std::ostringstream  str;

    if (rs<0) rs==0;
    if (re<0) re==0;

    if (rs>0 || re>0) {
	if (rs==re)  str << rs;
	else
	    if (rs==0 && re!=0)      str << ":" << re;
	    else
                str << rs << ":" << re;
    }
    return str.str();
}

string NATCompiler_ipt::PrintRule::_printTPorts(int rs,int re)
{
    std::ostringstream  str;

    if (rs<0) rs==0;
    if (re<0) re==0;

    if (rs>0 || re>0) {
	if (rs==re)  str << rs;
	else
	    if (rs==0 && re!=0)      str << "-" << re;
	    else
                str << rs << "-" << re;
    }
    return str.str();
}

string NATCompiler_ipt::PrintRule::_printICMP(ICMPService *srv)
{
    std::ostringstream  str;
    if (ICMPService::isA(srv) && srv->getInt("type")!=-1) {
	str << srv->getStr("type");
	if (srv->getInt("code")!=-1) 
	    str << "/" << srv->getStr("code") << " ";
    }
    return str.str();
}

string NATCompiler_ipt::PrintRule::_printIP(IPService *srv)
{
    std::ostringstream  str;
    if (IPService::isA(srv) ) {
	if (srv->getBool("fragm") || srv->getBool("short_fragm"))
	    str << "-f ";

        if  (srv->getBool("lsrr") ||
             srv->getBool("ssrr") ||
             srv->getBool("rr") ||
             srv->getBool("ts") ) str << " -m ipv4options ";

        if  (srv->getBool("lsrr")) str << " --lsrr";
        if  (srv->getBool("ssrr")) str << " --ssrr";
        if  (srv->getBool("rr"))   str << " --rr";
        if  (srv->getBool("ts"))   str << " --ts";
    }
    return str.str();
}

string NATCompiler_ipt::PrintRule::_printSrcPorts(Service *srv)
{
    std::ostringstream  str;
    if (TCPService::isA(srv) || UDPService::isA(srv)) {
	int rs=srv->getInt("src_range_start");
	int re=srv->getInt("src_range_end");
	str << _printOPorts(rs,re);
    }
    return str.str();
}

string NATCompiler_ipt::PrintRule::_printDstPorts(Service *srv)
{
    std::ostringstream  str;
    if (TCPService::isA(srv) || UDPService::isA(srv)) {
	int rs=srv->getInt("dst_range_start");
	int re=srv->getInt("dst_range_end");
	str << _printOPorts(rs,re);
    }
    return str.str();
}

string NATCompiler_ipt::PrintRule::_printSNATPorts(Service *srv)
{
    std::ostringstream  str;
    if (TCPService::isA(srv) || UDPService::isA(srv)) {
	int rs=srv->getInt("src_range_start");
	int re=srv->getInt("src_range_end");
	str << _printTPorts(rs,re);
    }
    return str.str();
}

string NATCompiler_ipt::PrintRule::_printDNATPorts(Service *srv)
{
    std::ostringstream  str;
    if (TCPService::isA(srv) || UDPService::isA(srv)) {
	int rs=srv->getInt("dst_range_start");
	int re=srv->getInt("dst_range_end");
	str << _printTPorts(rs,re);
    }
    return str.str();
}

/*
 * we made sure that all services in rel  represent the same protocol
 */
void NATCompiler_ipt::PrintRule::_printSrcService(RuleElementOSrv  *rel)
{
    Service *srv= rel->getFirst(true);

    if (rel->size()==1) {
	if (UDPService::isA(srv) || TCPService::isA(srv)) {
	    string str=_printSrcPorts( srv );
	    if (! str.empty() ) compiler->output << "--source-port " << str << " ";
	}
    } else {
/* use multiport */

	string str;
	bool  first=true;
	for (FWObject::iterator i=rel->begin(); i!=rel->end(); i++) {
	    FWObject *o= *i;
	    if (FWReference::cast(o)!=NULL) o=compiler->getCachedObj(o->getStr("ref"));

	    Service *s=Service::cast( o );
	    assert(s);
	    if (UDPService::isA(srv) || TCPService::isA(srv)) {
		if (!first) str+=",";
		str+= _printSrcPorts( s );
		if (!str.empty()) first=false;
	    }
	}
	if ( !str.empty() ) {
            if (compiler->fw->getStr("version")=="ge_1.2.6")
                compiler->output << "--source-ports ";
            else
                compiler->output << "--source-port ";
	    compiler->output << str << " ";
	}
    }
}

void NATCompiler_ipt::PrintRule::_printDstService(RuleElementOSrv  *rel)
{
    Service *srv= rel->getFirst(true);

    if (rel->size()==1) {
	if (UDPService::isA(srv) || TCPService::isA(srv)) {
	    string str=_printDstPorts( srv );
	    if (! str.empty() ) compiler->output << "--destination-port " << str << " ";
	}
	if (ICMPService::isA(srv)) {
	    string str=_printICMP( ICMPService::cast(srv) );
	    if (! str.empty() ) compiler->output << "--icmp-type " << str << " ";
	}
	if (IPService::isA(srv)) {
	    string str=_printIP( IPService::cast(srv) );
	    if (! str.empty() ) compiler->output << str << " ";
	}

    } else {
/* use multiport */

	string str;
	bool   first=true;
	for (FWObject::iterator i=rel->begin(); i!=rel->end(); i++) {
	    FWObject *o= *i;
	    if (FWReference::cast(o)!=NULL) o=compiler->getCachedObj(o->getStr("ref"));

	    Service *s=Service::cast( o );
	    assert(s);
	    if (UDPService::isA(srv) || TCPService::isA(srv)) {
		if (!first) str+=",";
		str+= _printDstPorts( s );
		if (!str.empty()) first=false;
	    }
	}
	if ( !str.empty() ) {
            if (compiler->fw->getStr("version")=="ge_1.2.6")
                compiler->output << "--destination-ports ";
            else
                compiler->output << "--destination-port ";
	    compiler->output << str << " ";
	}
    }
}

void NATCompiler_ipt::PrintRule::_printAddr(Address  *o,bool print_mask,bool print_range)
{
    IPAddress addr=o->getAddress();
    Netmask   mask=o->getNetmask();

    if (Interface::cast(o)!=NULL) 
    {
        if (Interface::cast(o)->isDyn()) return;
	mask=Netmask("255.255.255.255");
    }

    if (IPv4::cast(o)!=NULL) 
    {
	mask=Netmask("255.255.255.255");
    }

    if (print_range && AddressRange::cast(o)!=NULL) {
	IPAddress a1=AddressRange::cast(o)->getRangeStart();
	IPAddress a2=AddressRange::cast(o)->getRangeEnd();
	compiler->output << a1.toString() << "-" << a2.toString();
    } else {
	if (addr.toString()=="0.0.0.0" && mask.toString()=="0.0.0.0") {
	    compiler->output << "0/0";
	} else {	
	    compiler->output << addr.toString();
	    if (print_mask && mask.toString()!="255.255.255.255") {
		compiler->output << "/" << mask.getLength();
	    }
	}
    }
}


NATCompiler_ipt::PrintRule::PrintRule(const std::string &name) : 
    NATRuleProcessor(name)
{
    init=true; 

    chains["POSTROUTING"]  =true;
    chains["PREROUTING"]   =true;
    chains["SNAT"]         =true;
    chains["DNAT"]         =true;
    chains["MASQUERADE"]   =true;
    chains["REDIRECT"]     =true;
    chains["NETMAP"]       =true;
    chains["LOG"]          =true;
    chains["MARK"]         =true;
    chains["ACCEPT"]       =true;
    chains["REJECT"]       =true;
    chains["DROP"]         =true;
    chains["RETURN"]       =true;
    chains["OUTPUT"]       =true;
}

bool NATCompiler_ipt::PrintRule::processNext()
{
    NATRule *rule=getNext(); 
    if (rule==NULL) return false;

    tmp_queue.push_back(rule);

    string rl=rule->getLabel();
    if (rl!=current_rule_label) {
        compiler->output << "#" << endl;
        compiler->output << "#  Rule " << rl << endl;
        compiler->output << "# " << endl;

        string    comm=rule->getComment();
        string::size_type c1,c2;
        c1=0;
        while ( (c2=comm.find('\n',c1))!=string::npos ) {
            compiler->output << "# " << comm.substr(c1,c2-c1) << endl;
            c1=c2+1;
        }
        compiler->output << "# " << comm.substr(c1) << endl;
//        compiler->output << "# " << endl;

        current_rule_label=rl;
    }

/*
 *  check and create new chain if needed
 */
    string s;

    s=rule->getStr("ipt_chain");
    if ( ! chains[s] ) {
	compiler->output << "$IPTABLES -t nat -N " << s << endl;
	chains[s]=true;
    }

    s=rule->getStr("ipt_target");
    if ( ! chains[s] ) {
	compiler->output << "$IPTABLES -t nat -N " << s << endl;
	chains[s]=true;
    }



    RuleElementOSrc *osrcrel=rule->getOSrc();
    Address  *osrc=osrcrel->getFirst(true);  assert(osrc);
    RuleElementODst *odstrel=rule->getODst();
    Address  *odst=odstrel->getFirst(true);  assert(odst);
    RuleElementOSrv *osrvrel=rule->getOSrv();
    Service  *osrv=osrvrel->getFirst(true);  assert(osrv);

    Address  *tsrc=rule->getTSrc()->getFirst(true);  assert(tsrc);
    Address  *tdst=rule->getTDst()->getFirst(true);  assert(tdst);
    Service  *tsrv=rule->getTSrv()->getFirst(true);  assert(tsrv);

    Interface *iface=
	Interface::cast( rule->getRoot()->getById(rule->getInterfaceId() ,true) );


    compiler->output  << "$IPTABLES -t nat ";
    _printChainDirectionAndInterface(rule);

    if (! osrv->isAny() )
	_printProtocol(osrv);

    _printMultiport(rule);

    if (!osrc->isAny()) {
	compiler->output << " -s ";
	_printAddr(osrc);
    }

    if (!osrv->isAny()) {
	compiler->output << " ";
	_printSrcService(osrvrel);
    }

    if (!odst->isAny()) {
	compiler->output << " -d ";
	_printAddr(odst);
    }

    compiler->output << " ";
    _printDstService(osrvrel);

    compiler->output << "-j " << rule->getStr("ipt_target") << " ";

    switch (rule->getRuleType()) {
    case NATRule::SNAT:  
	if (rule->getStr("ipt_target")=="SNAT") {
	    compiler->output << "--to-source ";
	    _printAddr(tsrc,false,true);
	    string ports=_printSNATPorts(tsrv);
	    if (!ports.empty()) compiler->output << ":" << ports;
	}
	break;

    case NATRule::DNAT:  
	if (rule->getStr("ipt_target")=="DNAT") {
	    compiler->output << "--to-destination ";
	    _printAddr(tdst,false,true);
	    string ports=_printDNATPorts(tsrv);
	    if (!ports.empty()) compiler->output << ":" << ports;
	}
	break;

    case NATRule::SNetnat:
        compiler->output << "--to ";
        _printAddr(tsrc,true,false);
        break;

    case NATRule::DNetnat:
        compiler->output << "--to ";
        _printAddr(tdst,true,false);
        break;

    case NATRule::Redirect: 
	if (rule->getStr("ipt_target")=="REDIRECT") {
	    string ports=_printDNATPorts(tsrv);
	    if (!ports.empty()) compiler->output << "--to-ports " << ports;
	}
	break;
    default:	break;
    }
    compiler->output << " ";


    compiler->output  << endl;

    return true;
}

