/* 

                          Firewall Builder

                 Copyright (C) 2002 NetCitadel, LLC

  Author:  Vadim Kurland     vadim@vk.crocodile.org

  $Id: OSConfigurator_solaris.cc,v 1.14 2002/12/04 09:13:58 vkurland Exp $

  This program is free software which we release under the GNU General Public
  License. You may redistribute and/or modify this program under the terms
  of that license as published by the Free Software Foundation; either
  version 2 of the License, or (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
 
  To get a copy of the GNU General Public License, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

*/

#include <assert.h>

#include "OSConfigurator_solaris.hh"

#include "fwbuilder/Firewall.hh"
#include "fwbuilder/FWOptions.hh"
#include "fwbuilder/Interface.hh"
#include "fwbuilder/IPv4.hh"

#include <algorithm>
#include <functional>
#include <iostream>

using namespace libfwbuilder;
using namespace fwcompiler;
using namespace std;

string OSConfigurator_solaris::myPlatformName() { return "Solaris"; }

void OSConfigurator_solaris::processFirewallOptions() 
{
    FWOptions* options=fw->getOptionsObject();
    string s;

    s=options->getStr("solaris_ip_forward");
    if (!s.empty()) {
        if (s=="1" || s=="On" || s=="on") s="1";
        else                              s="0";

	output << "ndd -set /dev/ip ip_forwarding " << s << endl;
    }

    s=options->getStr("solaris_ip_ignore_redirect");
    if (!s.empty()) {
	if (s!="0" && s!="1")
	    throw FWException(_("Illegal value for OS parameter solaris_ip_ignore_redirect: '")+s+"'");

	output << "ndd -set /dev/ip ip_ignore_redirect " << s << endl;
    }

    s=options->getStr("solaris_ip_respond_to_echo_broadcast");
    if (!s.empty()) {
	if (s!="0" && s!="1")
	    throw FWException(_("Illegal value for OS parameter solaris_ip_respond_to_echo_broadcast: '")+s+"'");

	output << "ndd -set /dev/ip ip_respond_to_echo_broadcast " << s << endl;
    }

    s=options->getStr("solaris_ip_forward_directed_broadcasts");
    if (!s.empty()) {
	if (s!="0" && s!="1")
	    throw FWException(_("Illegal value for OS parameter solaris_ip_forward_directed_broadcasts: '")+s+"'");

	output << "ndd -set /dev/ip ip_forward_directed_broadcasts " << s << endl;
    }

    s=options->getStr("solaris_ip_forward_src_routed");
    if (!s.empty()) {
	if (s!="0" && s!="1")
	    throw FWException(_("Illegal value for OS parameter solaris_ip_forward_src_routed: '")+s+"'");

	output << "ndd -set /dev/ip ip_forward_src_routed " << s << endl;
    }

}

void OSConfigurator_solaris::addVirtualAddressForNAT(const Network &nw)
{
}

void OSConfigurator_solaris::addVirtualAddressForNAT(const IPAddress &addr)
{
    if (virtual_addresses.empty() || 
	find(virtual_addresses.begin(),virtual_addresses.end(),addr)==virtual_addresses.end()) {

        FWObjectTypedChildIterator i=fw->findByType(Interface::TYPENAME);
        for ( ; i!=i.end(); ++i ) {
	    Interface *iface=dynamic_cast<Interface*>(*i);
	    assert(iface);

	    FWObjectTypedChildIterator j=iface->findByType(IPv4::TYPENAME);
	    for ( ; j!=j.end(); ++j ) {
		IPv4 *iaddr=IPv4::cast(*j);

                IPNetwork n( iaddr->getAddress() , iaddr->getNetmask() );
                if ( n.belongs( addr ) ) {
                    output << "ifconfig " 
                           << iface->getName() << " "
                           << addr.toString() << " alias" << endl;
                    virtual_addresses.push_back(addr);
                    return;
                }
            }
	}
	throw FWException(_("Can not add virtual address ") + addr.toString() );
    }

} 
   
int OSConfigurator_solaris::prolog()
{
    FWOptions* options=fw->getOptionsObject();
    string firewall_dir=options->getStr("firewall_dir");
    if (firewall_dir=="") firewall_dir="/etc";

    char          *timestr;
    time_t         tm;
    struct tm     *stm;

    tm=time(NULL);
    stm=localtime(&tm);
    timestr=strdup(ctime(&tm));
    timestr[ strlen(timestr)-1 ]='\0';
    
    char* user_name=(char*)g_get_user_name();
    if (user_name==NULL) {
        cerr << _("Can't figure out your user name, aborting") << endl;
        exit(1);
    }

    output << _("#\n\
#  This is automatically generated file. DO NOT MODIFY !\n\
#\n\
#  Firewall Builder  fwb_ipf v") << VERSION << RELEASE_NUM << _(" \n\
#\n\
#  Generated ") << timestr << " " << tzname[stm->tm_isdst] << _(" by ") 
            << user_name << "\n#\n#\n#\n#\n";

    output << "cd " << firewall_dir << " || exit 1" << endl << endl;

    printPathForAllTools("solaris");

    output << "log \"";
    output << _("Activating firewall script generated ")
           << timestr << " " << tzname[stm->tm_isdst] << _(" by ")
           << user_name;
    output << "\"" << endl;


    output << endl 
           << "$IPF -Fa" << endl
           << "$IPNAT -C" << endl;

    output << endl;

    processFirewallOptions();
    return 0;
}

void  OSConfigurator_solaris::printPathForAllTools(const string &os)
{
    FWOptions* options=fw->getOptionsObject();
    
    string s, path_ipf, path_ipnat, path_logger;

    s=options->getStr("solaris_path_ipf");
    if (!s.empty()) path_ipf=s;
    else            path_ipf=os_data.getPathForTool(os,OSData::IPF);

    s=options->getStr("solaris_path_ipnat");
    if (!s.empty()) path_ipnat=s;
    else            path_ipnat=os_data.getPathForTool(os,OSData::IPNAT);

    s=options->getStr("solaris_path_logger");
    if (!s.empty()) path_logger=s;
    else            path_logger=os_data.getPathForTool(os,OSData::LOGGER);


    output << "check() {" << endl;
    output << "  if test ! -x \"$1\"; then" << endl;
    output << "    echo \"$1 " << _("not found or is not executable") << "\"" << endl;
    output << "    exit 1" << endl;
    output << "  fi" << endl;
    output << "}" << endl;
    output << endl;
    output << "log() {" << endl;
    output << "  if test -x \"$LOGGER\"; then" << endl;
    output << "    logger -p info \"$1\"" << endl;
    output << "  fi" << endl;
    output << "}" << endl;
    output << endl;

    output << "IPF=\""    + path_ipf    + "\"\n";
    output << "IPNAT=\""  + path_ipnat  + "\"\n";
    output << "LOGGER=\"" + path_logger + "\"\n";
    output << endl;

    output << "check $IPF" << endl;
    output << "check $IPNAT" << endl;

    output << endl;
}



