/* 

                          Firewall Builder

                 Copyright (C) 2000 NetCitadel, LLC

  Author:  Vadim Kurland     vadim@vk.crocodile.org

  $Id: TargetDialog_pf.cc,v 1.8 2002/11/29 03:07:33 vkurland Exp $

  This program is free software which we release under the GNU General Public
  License. You may redistribute and/or modify this program under the terms
  of that license as published by the Free Software Foundation; either
  version 2 of the License, or (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
 
  To get a copy of the GNU General Public License, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

*/


#include "config.h"
#include "fwbuilder/Firewall.hh"

#include "TargetDialog_pf.hh"
#include "fwbuilder/Resources.hh"

using namespace libfwbuilder;

DialogPlugin* create_dialog_pf(FWObject *o)
{
    return new TargetDialog_pf(o);
}


TargetDialog_pf::TargetDialog_pf(FWObject *obj) 
{ 
    object=obj;

    fwoptions=(Firewall::cast(object))->getOptionsObject();
    assert(fwoptions!=NULL);


    label556->set_line_wrap(true);

    ((Gtk::Label*)(pf_no_optimisation->get_child()))->set_alignment(0.0,0.5);
    ((Gtk::Label*)(pf_no_optimisation->get_child()))->set_justify(GTK_JUSTIFY_LEFT);

    ((Gtk::Label*)(pf_modulate_state->get_child()))->set_alignment(0.0,0.5);
    ((Gtk::Label*)(pf_modulate_state->get_child()))->set_justify(GTK_JUSTIFY_LEFT);

    ((Gtk::Label*)(pf_manage_virtual_addr->get_child()))->set_alignment(0.0,0.5);
    ((Gtk::Label*)(pf_manage_virtual_addr->get_child()))->set_justify(GTK_JUSTIFY_LEFT);

    ((Gtk::Label*)(pf_debug->get_child()))->set_alignment(0.0,0.5);

    label570->set_line_wrap(true);

    ((Gtk::Label*)(pf_in_out_code->get_child()))->set_alignment(0.0,0.5);
    ((Gtk::Label*)(pf_pass_all_out->get_child()))->set_alignment(0.0,0.5);

    options.push_back( DialogOption( pf_in_out_code      , "in_out_code"   ));
    options.push_back( DialogOption( pf_pass_all_out     , "pass_all_out"  ));
    
    options.push_back( DialogOption( pf_check_shading    , "check_shading" ));
    options.push_back( DialogOption( pf_ignore_empty_groups, "ignore_empty_groups" ));
    
    options.push_back( DialogOption( pf_no_optimisation  , "no_optimisation"  ));
    options.push_back( DialogOption( pf_modulate_state   , "modulate_state"   ));
    
    options.push_back( DialogOption( pf_debug                            , "debug"           ));
    options.push_back( DialogOption( pf_manage_virtual_addr              , "manage_virtual_addr"     ));
    options.push_back( DialogOption( pf_fw_dir                           , "firewall_dir"            ));
    
}

void TargetDialog_pf::on_changed()
{   
    DialogPlugin::data_changed_flag(true);
}

void TargetDialog_pf::readDefaults()
{
    pf_fw_dir->set_text(
        Resources::platform_res["pf"]->getResourceStr(
            "/FWBuilderResources/Target/options/firewall_dir"));
}

void TargetDialog_pf::wrk2dlg()
{
    map<string,string>        fw_versions_menu;

    fw_versions_menu.clear();

    fw_versions_menu[DialogPlugin::magic_empty_option] = "";

    fw_version->set_menu(fw_versions_menu);
    fw_version->changed.connect(SigC::slot(this,&TargetDialog_pf::on_changed));
    fw_version->show_all();

    loadOptionMenu (object,fw_version,"version");

    loadPageOptions(fwoptions,  options);
}

bool TargetDialog_pf::dlg2wrk()
{
    saveOptionMenu(object,fw_version,"version");

    savePageOptions(fwoptions,  options );
    return true;
}
